"""
CLI entry point
"""
import logging
from datetime import datetime
from decimal import Decimal

import click
import click_log

from .app import PriceDbApplication
from .map_cli import symbol_map
from .model import PriceModel, SecuritySymbol

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


@click.group()
@click_log.simple_verbosity_option(logger)
def cli():
    """ entry point """
    pass


@click.command()
@click.option("--symbol", prompt=True, type=str)
@click.option("--date", prompt=True)
@click.option("--value", prompt=True)
@click.option("--currency", prompt=True, type=str)
def add(symbol: str, date, value, currency: str):
    """ Add individual price """
    app = PriceDbApplication()
    price = PriceModel()

    security = SecuritySymbol("", "")
    security.parse(symbol)
    price.symbol.mnemonic = price.symbol.mnemonic.upper()

    date_str = f"{date}"
    date_format = "%Y-%m-%d"
    # if time:
    #     date_str = f"{date_str}T{time}"
    #     date_format += "T%H:%M:%S"
    price.datetime = datetime.strptime(date_str, date_format)

    price.value = Decimal(value)
    price.currency = currency.upper()
    app.add_price(price)
    app.save()

    click.echo("Price added.")


@click.command("import")
@click.argument("file", "FILE - path to the .csv file to import")
@click.argument("currency", "currency - to use for imported prices")
@click_log.simple_verbosity_option(logger)
def import_csv(file, currency: str):
    """ Import prices from CSV file """
    logger.debug(f"currency = {currency}")
    # auto-convert to uppercase.
    currency = currency.upper()

    app = PriceDbApplication()
    app.logger = logger
    app.import_prices(file, currency)


@click.command()
@click.argument("symbol")
def last(symbol: str):
    """ displays last price for symbol """
    # convert to uppercase
    symbol = symbol.upper()
    # extract namespace
    sec_symbol = SecuritySymbol("", "")
    sec_symbol.parse(symbol)

    app = PriceDbApplication()
    latest = app.get_latest_price(sec_symbol)
    assert isinstance(latest, PriceModel)
    print(f"{latest}")


@click.command("list")
@click.option("--date", help="The date for which to show prices.")
@click.option("--currency", help="The currency for which to show prices.")
@click.option("--last", is_flag=True, help="Display only the latest prices per symbol")
@click_log.simple_verbosity_option(logger)
def list_prices(date, currency, last):
    """ Display all prices """
    app = PriceDbApplication()
    app.logger = logger

    if last:
        # fetch only the last prices
        prices = app.get_latest_prices()
    else:
        prices = app.get_prices(date, currency)
    for price in prices:
        print(price)

    print(f"{len(prices)} records found.")


@click.command("dl")
@click.option("--symbol", "-s", help="Symbol for the individual price to download")
@click.option("--file", "-f", help="The text file containing the symbols to download")
@click.option("--agent", "-a", help="Agent to use for download (vanguard_au, morningstar, alphavantage)", required=True)
@click.option("--currency", "-c", help="Currency symbol to use for the price(s)", required=True)
@click_log.simple_verbosity_option(logger)
def download(symbol: str, file: str, agent: str, currency: str):
    """ Download the latest prices """
    app = PriceDbApplication()
    app.logger = logger

    # try:
    #     import finance_quote
    # except ImportError as error:
    #     app.logger.error(error)
    #     print("You need to install finance_quote package for this functionality.")

    if symbol:
        # download individual price
        app.download_price(symbol, currency, agent)
        return

    if file:
        # Download prices from the file. One symbol per line.
        app.download_prices_from_file(file, currency, agent)
        return

    # download all prices
    print("Please use --symbol or --file option. --help for more info. Symbol will have preference over file.")


@click.command("prune")
@click.option("--all", "-a", is_flag=True, help="Delete historical prices for all symbols")
@click.option("--symbol", "-s", help="Symbol for which to delete prices")
@click_log.simple_verbosity_option(logger)
def prune(symbol: str, all):
    """ Delete old prices, leaving just the last. """
    app = PriceDbApplication()
    app.logger = logger

    if symbol is not None:
        sec_symbol = SecuritySymbol("", "")
        sec_symbol.parse(symbol)

        app.prune(sec_symbol)
    else:
        app.prune_all()


######
cli.add_command(add)
cli.add_command(download)
cli.add_command(import_csv)
cli.add_command(symbol_map)
cli.add_command(last)
cli.add_command(list_prices)
cli.add_command(prune)

##################################################
# Debug run
# if __name__ == "__main__":
#     import_csv(["file", "data/AUD.csv"])



"""
Price downloaders and parsers.
These can be used as the first aid, to populate the price database but are not
a comprehensive solution.
"""
import logging
from enum import Enum, auto

from pricedb import SecuritySymbol

from .alphavantage import AlphaVantageDownloader
from .morningstar import MorningstarDownloader
from .vanguard_au import VanguardAuDownloader
from .fixerio import Fixerio


class DownloadAgents(Enum):
    """ Available agents for price download """
    alphavantage = auto(),
    fixerio = auto(),
    morningstar = auto(),
    vanguard_au = auto()


class PriceDownloader:
    """ Proxy class for downloading prices """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def download(self, symbol: str, currency: str = None, agent: str = None):
        """ Download single latest price """
        assert agent is not None
        assert isinstance(agent, str)

        symbol = symbol.upper()
        currency = currency.upper()
        agent = agent.lower()
        actor = None
        price = None

        security_symbol = SecuritySymbol("", "")
        security_symbol.parse(symbol)

        if agent == DownloadAgents.morningstar.name:
            actor = MorningstarDownloader()
        elif agent == DownloadAgents.vanguard_au.name:
            actor = VanguardAuDownloader()
        elif agent == DownloadAgents.alphavantage.name:
            actor = AlphaVantageDownloader()
        elif agent == DownloadAgents.fixerio.name:
            actor = Fixerio()
        else:
            raise ValueError("No agent specified for price download.")

        if actor:
            actor.logger = self.logger
            price = actor.download(security_symbol, currency)

        return price



""" main api point """
import logging
from decimal import Decimal
from typing import List

from . import dal, mappers, utils
from .csv import CsvParser
from .dal import Price
from .download import PriceDownloader
from .model import PriceModel, SecuritySymbol
from .repositories import PriceRepository


class PriceDbApplication:
    """ Contains the main public interfaces """

    def __init__(self, session=None):
        self.logger = logging.getLogger(__name__)
        self.price_repo = None
        self.__session = session

    def add_price(self, price: PriceModel):
        """ Creates a new price record """
        # assert isinstance(price, PriceModel)

        if not price:
            raise ValueError("Cannot add price. The received model is null!")

        mapper = mappers.PriceMapper()
        entity = mapper.map_model(price)

        self.add_price_entity(entity)

    def add_price_entity(self, price: dal.Price):
        """ Adds the price """
        # check if the price already exists in db.
        repo = self.get_price_repository()
        existing = (
            repo.query
                .filter(dal.Price.namespace == price.namespace)
                .filter(dal.Price.symbol == price.symbol)
                .filter(dal.Price.date == price.date)
                .filter(dal.Price.time == price.time)
                .first()
        )
        if existing:
            # Update existing price.
            new_value = Decimal(price.value) / Decimal(price.denom)
            self.logger.info(f"Exists: {price}")
            if price.currency != existing.currency:
                raise ValueError(
                    f"The currency is different for price {price}!")
            if existing.value != price.value:
                existing.value = price.value
                self.logger.info(f"Updating to {new_value}.")
            if existing.denom != price.denom:
                existing.denom = price.denom
        else:
            # Insert new price
            self.session.add(price)
            self.logger.info(f"Added {price}")

    def download_price(self, symbol: str, currency: str, agent: str) -> PriceModel:
        """ Download and save price online """
        price = self.__download_price(symbol, currency, agent)
        self.save()
        return price

    def download_prices_from_file(self, file_path: str, currency: str, agent: str):
        """ Reads price symbols from a file and downloads prices """
        # read symbols from a text file
        symbols = utils.read_lines_from_file(file_path)
        for symbol in symbols:
            self.__download_price(symbol.strip(), currency, agent)
        self.save()

    def import_prices(self, file_path: str, currency_symbol: str):
        """ Incomplete """
        assert isinstance(file_path, str)
        assert isinstance(currency_symbol, str)

        self.logger.debug(f"Importing {file_path}")
        parser = CsvParser()
        prices = parser.parse_file(file_path, currency_symbol)

        counter = 0
        session = self.session
        # Create insert statements
        mapper = mappers.PriceMapper()
        for price in prices:
            new_price = mapper.map_model(price)
            self.add_price_entity(new_price)
            counter += 1
        # Save all to database
        session.commit()
        print(f"{counter} records inserted.")

    def get_latest_price(self, symbol: SecuritySymbol) -> PriceModel:
        """ Returns the latest price for the given symbol """
        # TODO should include the currency? Need a public model for exposing the result.
        assert isinstance(symbol, SecuritySymbol)

        session = self.session
        repo = PriceRepository(session)
        query = (
            repo.query
                .filter(dal.Price.symbol == symbol.mnemonic)
                .order_by(dal.Price.date.desc(), dal.Price.time.desc())
        )
        if symbol.namespace:
            query = query.filter(dal.Price.namespace == symbol.namespace)

        latest = query.first()

        # map
        mapper = mappers.PriceMapper()
        result = mapper.map_entity(latest)

        return result

    @property
    def session(self):
        """ Returns the current db session """
        if not self.__session:
            self.__session = dal.get_default_session()
        return self.__session

    def get_prices(self, date: str, currency: str) -> List[PriceModel]:
        """ Fetches all the prices for the given arguments """
        session = self.session
        repo = PriceRepository(session)
        query = repo.query
        if date:
            query = query.filter(dal.Price.date == date)
        if currency:
            query = query.filter(dal.Price.currency == currency)
        # Sort by symbol.
        query = query.order_by(dal.Price.namespace, dal.Price.symbol)
        price_entities = query.all()

        mapper = mappers.PriceMapper()
        result = []
        for entity in price_entities:
            model = mapper.map_entity(entity)
            result.append(model)
        return result

    def get_prices_on(self, on_date: str, namespace: str, symbol: str):
        """ Returns the latest price on the date """
        repo = self.get_price_repository()
        query = (
            repo.query.filter(dal.Price.namespace == namespace)
                .filter(dal.Price.symbol == symbol)
                .filter(dal.Price.date == on_date)
                .order_by(dal.Price.time.desc())
        )
        result = query.first()
        # logging.debug(result)
        return result

    def get_latest_prices(self):
        """ Fetches the latest prices for all symbols """
        # get all symbols first, for which we have prices available
        repo = PriceRepository()

        query = (
            repo.session.query(dal.Price.namespace, dal.Price.symbol)
                .order_by(dal.Price.namespace, dal.Price.symbol)
                .distinct()
        )
        all_symbols = query.all()
        # self.logger.debug(all_symbols)

        # Get the latest prices for these symbols
        latest_prices = []
        for symbol_price in all_symbols:
            symbol = SecuritySymbol(symbol_price.namespace, symbol_price.symbol)
            latest = self.get_latest_price(symbol)
            latest_prices.append(latest)
        return latest_prices

    def get_price_repository(self):
        """ Price repository """
        if not self.price_repo:
            self.price_repo = PriceRepository(self.session)
        return self.price_repo

    def prune_all(self):
        """ Prune historical prices for all symbols, leaving only the latest """
        # get all symbols that have prices
        repo = PriceRepository()
        items = repo.query.distinct(Price.namespace, Price.symbol).all()
        # self.logger.debug(items)

        for item in items:
            symbol = SecuritySymbol(item.namespace, item.symbol)
            self.prune(symbol)

    def prune(self, symbol: SecuritySymbol):
        """ Delete all but the latest available price for the given symbol """
        assert isinstance(symbol, SecuritySymbol)

        self.logger.debug(f"pruning prices for {symbol}")

        repo = PriceRepository()
        query = (
            repo.query.filter(Price.namespace == symbol.namespace)
            .filter(Price.symbol == symbol.mnemonic)
            .order_by(Price.date.desc())
            .order_by(Price.time.desc())
        )
        all_prices = query.all()
        # self.logger.debug(f"fetched {all_prices}")

        first = True
        for single in all_prices:
            if not first:
                repo.query.filter(Price.id == single.id).delete()
                self.logger.debug(f"deleting {single.id}")
            else:
                first = False

        repo.save()

    def save(self):
        """ Save changes """
        if self.__session:
            self.session.commit()
        else:
            self.logger.warning(f"Save called but no session open.")

    def __download_price(self, symbol: str, currency: str, agent: str):
        """ Downloads and parses the price """
        if not symbol:
            return None

        dl = PriceDownloader()
        dl.logger = self.logger

        price = dl.download(symbol, currency, agent)
        if not price:
            raise ValueError(f"Price not downloaded/parsed for {symbol}.")

        self.add_price(price)

        return price



""" Mapping entities to domain model objects """
from datetime import datetime
from decimal import Decimal
from . import dal
from .dal import Price
from .model import PriceModel, SecuritySymbol


class PriceMapper:
    """ Map price entity """
    def __init__(self):
        pass

    def map_entity(self, entity: dal.Price) -> PriceModel:
        """ Map the price entity """
        if not entity:
            return None

        result = PriceModel()
        result.currency = entity.currency

        # date/time
        dt_string = entity.date
        format_string = "%Y-%m-%d"
        if entity.time:
            dt_string += f"T{entity.time}"
            format_string += "T%H:%M:%S"
        result.datetime = datetime.strptime(dt_string, format_string)
        assert isinstance(result.datetime, datetime)

        result.namespace = entity.namespace
        result.symbol = entity.symbol
        # Value
        value = Decimal(entity.value) / Decimal(entity.denom)
        result.value = Decimal(value)

        return result

    def map_model(self, model: PriceModel) -> Price:
        """ Parse into the Price entity, ready for saving """
        # assert isinstance(model, PriceModel)
        assert isinstance(model.symbol, SecuritySymbol)
        assert isinstance(model.datetime, datetime)

        entity = Price()

        # Format date as ISO string
        date_iso = f"{model.datetime.year}-{model.datetime.month:02d}-{model.datetime.day:02d}"
        entity.date = date_iso

        entity.time = f"{model.datetime.hour:02d}:{model.datetime.minute:02d}:{model.datetime.second:02d}"

        # Symbol
        # properly mapped symbols have a namespace, except for the US markets
        # TODO check this with .csv import
        if model.symbol.namespace:
            entity.namespace = model.symbol.namespace.upper()
        entity.symbol = model.symbol.mnemonic.upper()

        assert isinstance(model.value, Decimal)
        # Find number of decimal places
        dec_places = abs(model.value.as_tuple().exponent)
        entity.denom = 10 ** dec_places
        # Price value
        entity.value = int(model.value * entity.denom)

        # Currency
        entity.currency = model.currency.upper()

        # self.logger.debug(f"{entity}")
        return entity



""" CSV parsing elements """
from decimal import Decimal
from datetime import datetime
from typing import List
import logging
from . import dal # import get_session, Price, SymbolMap
from . import config
from .repositories import SymbolMapRepository
from .model import PriceModel
from .utils import read_lines_from_file

# class CsvPrice:
#     """ The price object parsed from a .csv file """
#     def __init__(self):
#         self.date: datetime = None
#         self.symbol: str = None
#         self.value: Decimal = None

#     def __repr__(self):
#         return f"<CsvPrice (date={self.date},symbol={self.symbol},value={self.value})>"

class CsvParser:
    """ Parse CSV file """
    def __init__(self):
        self.session = None
        self.symbol_maps = None

    def parse_file(self, file_path, currency) -> List[PriceModel]:
        """ Load and parse a .csv file """
        # load file
                # read csv into memory?
        contents = self.load_file(file_path)
        prices = []

        # parse price elements
        for line in contents:
            price = self.parse_line(line)
            assert isinstance(price, PriceModel)
            price.currency = currency
            prices.append(price)

        return prices

    def load_file(self, file_path) -> List[str]:
        """ Loads the content of the text file """
        content = []
        content = read_lines_from_file(file_path)
        return content

    def parse_line(self, line: str) -> PriceModel:
        """ Parse a CSV line into a price element """
        line = line.rstrip()
        parts = line.split(',')

        result = PriceModel()

        # symbol
        result.symbol = self.translate_symbol(parts[0])

        # value
        result.value = Decimal(parts[1])

        # date
        date_str = parts[2]
        date_str = date_str.replace('"', '')
        date_parts = date_str.split('/')

        year_str = date_parts[2]
        month_str = date_parts[1]
        day_str = date_parts[0]

        logging.debug(f"parsing {date_parts} into date")
        result.datetime = datetime(int(year_str), int(month_str), int(day_str))

        return result

    def translate_symbol(self, in_symbol: str) -> str:
        """ translate the incoming symbol into locally-used """
        # read all mappings from the db
        if not self.symbol_maps:
            self.__load_symbol_maps()
        # translate the incoming symbol
        result = self.symbol_maps[in_symbol] if in_symbol in self.symbol_maps else in_symbol

        return result

    def __load_symbol_maps(self):
        """ Loads all symbol maps from db """
        repo = SymbolMapRepository(self.__get_session())
        all_maps = repo.get_all()
        self.symbol_maps = {}
        for item in all_maps:
            self.symbol_maps[item.in_symbol] = item.out_symbol

    def __get_session(self):
        """ Reuses the same db session """
        if not self.session:
            self.session = dal.get_default_session()
        return self.session



""" Morningstar price downloader """
import logging
import urllib.parse
import urllib.request
from datetime import datetime
from decimal import Decimal, InvalidOperation

from bs4 import BeautifulSoup

from pricedb.model import PriceModel, SecuritySymbol


class MorningstarDownloader:
    """ Fetches prices from Morningstar site """
    def __init__(self):
        self.url = "http://quotes.morningstar.com/stockq/c-header"
        self.params = {'t': 'symbol'}
        self.namespaces = {
            "AMS": "XAMS",
            "ASX": "XASX",
            "XETRA": "XETR",
            "LSE": "XLON",
            "NASDAQ": "XNAS",
            "NYSE": "XNYS",
            "NYSEARCA": "ARCX"
        }
        self.logger = logging.getLogger(__name__)

    def download(self, symbol: SecuritySymbol, currency: str) -> PriceModel:
        """ Download the given symbol """
        if not symbol.namespace:
            raise ValueError(f"Namespace not sent for {symbol}")

        # get the translated namespace
        local_namespace = self.namespaces[symbol.namespace]
        self.params["t"] = f"{local_namespace}:{symbol.mnemonic}"
        # assemble url
        params = urllib.parse.urlencode(self.params)
        url = self.url + '?' + params
        self.logger.debug(f"fetching price from {url}")

        with urllib.request.urlopen(url) as response:
            html = response.read()

        if not html:
            return None

        # parse HTML
        price = self.parse_price(html)
        if price:
            price.symbol = symbol
        # compare currency
        if price.currency != currency:
            raise ValueError(f"Currency does not match for {symbol}! {currency}")

        return price

    def parse_price(self, page: str) -> PriceModel:
        """ parse html to get the price """
        result = PriceModel()
        soup = BeautifulSoup(page, 'html.parser')

        # Price value
        price_el = soup.find(id='last-price-value')
        if not price_el:
            logging.debug(f"Received from mstar: {page}")
            raise ValueError("No price info found in returned HTML.")

        value = price_el.get_text().strip()
        try:
            result.value = Decimal(value)
        except InvalidOperation:
            message = f"Could not parse {value}"
            print(message)
            self.logger.error(message)
            return None

        # The rest
        date_str = soup.find(id="asOfDate").get_text().strip()
        date_val = datetime.strptime(date_str, "%m/%d/%Y %H:%M:%S")
        result.datetime = date_val

        # tz_str = soup.find(id="timezone").get_text().strip()

        currency = soup.find(id="curency").get_text().strip()
        result.currency = currency

        return result



"""
The script downloads fund prices from Vanguard Australia site.
Retail Funds
Vanguard Diversified Bond Index Fund                  VAN0101AU   8123 = VANGUARD:BOND
Vanguard International Shares Index Fund (Hedged)     VAN0107AU   8146 = VANGUARD:HINT
Vanguard Australian Property Securities Index Fund    VAN0012AU   8147 = VANGUARD:PROP
Vanguard Australian Shares High Yield Fund            VAN0017AU   8148 = VANGUARD:HY
"""
import logging
from datetime import datetime
from decimal import Decimal

import requests

from pricedb.model import PriceModel, SecuritySymbol

try:
    import simplejson as json
except ImportError:
    import json


class FundInfo:
    """ Vg fund info. A DTO. """

    def __init__(self):
        self.name: str = None
        self.identifier = None
        self.date = None
        self.value: str = None
        self.mstar_id = None
        self.currency = None

    def __repr__(self):
        return f"<FundInfo (name='{self.name}',id='{self.identifier}',value={self.value})>"


class VanguardAuDownloader:
    """
    Downloads prices from Vanguard Australia
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        # cached HTML
        self.__data = None
        self.fund_map = {
            "VANGUARD:BOND": "8123",
            "VANGUARD:HINT": "8146",
            "VANGUARD:PROP": "8147",
            "VANGUARD:HY": "8148"
        }

    def download(self, symbol: SecuritySymbol, currency: str):
        """ Download price """
        if symbol.namespace != "Vanguard".upper():
            raise ValueError(f"Only Vanguard namespace is handled by this agent. Requested {symbol.namespace}:{symbol.mnemonic}!")

        fund_data = self.__load_fund_data()

        fund_id = self.fund_map[str(symbol)]
        fund_info = self.__get_fund_price(fund_data, fund_id)
        # self.logger.debug(f"{price}")

        result = PriceModel()

        date_format = "%d %b %Y"
        result.datetime = datetime.strptime(fund_info.date, date_format)

        result.symbol = SecuritySymbol("VANGUARD", symbol.mnemonic)

        value = fund_info.value.strip("$")
        result.value = Decimal(value)

        result.currency = "AUD"
        return result

    def get_fund_info(self, fund_symbol: str):
        """ For compatibility. Returns the full fund info """
        self.__load_fund_data()
        fund_id = self.fund_map[fund_symbol]
        if not fund_id:
            raise ValueError(f"Fund not found in map {fund_symbol}")

        return self.__get_fund_price(self.__data, fund_id)

    def __load_fund_data(self):
        """
        Fetches retail fund prices.
        """
        if self.__data:
            return self.__data

        # url = "https://www.vanguardinvestments.com.au/retail/ret/investments/product.html"
        # url = "https://www.vanguardinvestments.com.au/retail/mvc/getNavPrice?portId=" + fund_id
        # pylint: disable=C0301
        # url = "https://www.vanguardinvestments.com.au/retail/mvc/getNavPriceList.jsonp"
        # url = "https://intlgra-globm-209.gra.international.vgdynamic.info/rs/gre/gra/datasets/auw-retail-listview-data.jsonp"
        url = "https://intlgra-graapp-72-prd.gra.international.vgdynamic.info/rs/gre/gra/datasets/auw-retail-listview-data.jsonp"
        response = requests.get(url)
        if response.status_code != 200:
            return "Error"

        content = response.content
        content = str(content, 'utf-8')

        # clean-up the response
        if content.startswith("callback("):
            length = len(content) - 1
            content = content[9:length]

        # cache the downloaded page? The prices are daily.

        content_json = json.loads(content)

        self.__data = content_json["fundData"]
        return self.__data

    def __get_fund_price(self, fund_data, fund_id: str):
        """
        Extracts the price value from json response.
        Returns the Price object with name, identifier, date, value, mstar_id.
        """
        info_json = fund_data[fund_id]
        fund_info = FundInfo()

        fund_info.name = info_json["name"]
        fund_info.identifier = info_json["identifier"]
        fund_info.date = info_json["asOfDate"]
        # Using NAV price for the value.
        fund_info.value = info_json["navPrice"]
        fund_info.mstar_id = info_json["mStarID"]
        # Currency?
        # fund_info.currency = info_json["currency"]["currencyCode"]

        return fund_info
